/*
 * bobpong.c
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include "extern.h"
#include "ether.h"

/*
 * bobnet specific globals
 */
#define NDESC		64
#define MAX_SIZE	1048576

struct vi_cseg	descs[NDESC * 2];
char		packets[2][MAX_SIZE];

struct free_entry {
    struct vi_cseg	*desc;
    struct free_entry	*next;
} free_descs[NDESC];
struct free_entry *free_head;

int nic_handle, vi_handle, desc_mh, packet_mh;

const char *dev = "/dev/vi0";

void		bobnet_init(const char *peer_name);
void		bobnet_finish(void);
int		bobnet_send(void *packet, int, int);
int		bobnet_recv(void *packet, int, int);
void bobnet_send_immediate(u_int32_t data, int dest);
void bobnet_recv_immediate(u_int32_t *data, int dest);
int		bobnet_sendwait(struct vi_cseg **, int);
int		bobnet_recvwait(struct vi_cseg **, int);

void		sync_ping(void);
void		sync_pong(void);
void		async_ping(void);
void		async_pong(void);

#ifdef __linux__
# define bcopy(s,d,n) memcpy(d,s,n)
# if !(defined(__GLIBC__) && __GLIBC__ >= 2)
static void
timersub(struct timeval *t1, struct timeval *t2, struct timeval *res)
{
    res->tv_sec = t1->tv_sec - t2->tv_sec;
    if (t1->tv_usec > t2->tv_usec)
	res->tv_usec = t1->tv_usec - t2->tv_usec;
    else {
	--res->tv_sec;
	res->tv_usec = (1000000 + t1->tv_usec) - t2->tv_usec;
    }
}
# endif
#endif

static void
usage(void)
{
    printf("Usage: bobpong -t [-P <dvi>] [-n <number>] [-s <size>] <dest host>\n");
    printf("       bobpong -r [-P <dvi>]\n");
    exit(1);
}

int number = 1;	/* default # of pings */
int size = 64;	/* default packet size */
int listen_port = 1010;  /* default rendez-vous address */

static void
catch(int sig)
{
    printf("Caught signal %d\n", sig);
    bobnet_finish();
    exit(0);
}

int
main(int argc, char *argv[])
{
    int tflag, rflag, sync;
    int i;
    char *peer_name = NULL;

    tflag = rflag = 0;
    sync = 1;
    while ((i = getopt(argc, argv, "n:s:trP:")) != EOF)
	switch (i) {
	  case 't':
	    tflag = 1;
	    break;
	  case 'r':
	    rflag = 1;
	    break;
	  case 'n':
	    number = atoi(optarg);
	    break;
	  case 's':
	    size = atoi(optarg);
	    break;
	  case 'P':
	    listen_port = atoi(optarg);
	    break;
	  default:
	    usage();
	}

    argc -= optind;
    argv += optind;

    if (tflag + rflag != 1)
	usage();

    if (tflag) {
	if (!argc) usage();
	peer_name = *argv;
	--argc; ++argv;
    }

    if (argc) usage();

    if (size > MAX_SIZE) {
	fprintf(stderr, "Size %d too big, recompile?\n", size);
	return 1;
    }

    bobnet_init(peer_name);
    printf("connected\n");
    signal(SIGINT, catch);
    signal(SIGTERM, catch);
    
    /* send number/size to receiver */
    if (tflag) {
	bobnet_send_immediate(size, 0); sleep(1);
	bobnet_send_immediate(number, 0); sleep(1);
    } else {
	bobnet_recv_immediate(&size, 0);
	bobnet_recv_immediate(&number, 0);
    }

    if (tflag) {
	struct timeval before, after, elapsed;
	double t;

	/*
	 * Stick a pattern in the packet so it looks okay on tcpdump
	 */
	for (i=0; i<sizeof(packets[0]); i++)
	    packets[0][i] = i;
	
	gettimeofday(&before, NULL);
	if (sync)
	    sync_ping();
	else
	    async_ping();
	gettimeofday(&after, NULL);

	timersub(&after, &before, &elapsed);

	t = (double) elapsed.tv_sec + (double) elapsed.tv_usec * 1.0e-6;
	printf("%d at %d, %5.2f us, %5.2f Mb/s\n",
		number, size,
		1.0e6 * t / (double) (number * 2),
		1.0e-6 * (double)(16 * size * number) / t);
		/* 2-way, 8 bit/byte */
    }
    if (rflag) {
	for (i=0; i<sizeof(packets[1]); i++)
	    packets[1][i] = 0x80 + i;

	if (sync)
	    sync_pong();
	else
	    async_pong();
    }

    bobnet_finish();
    return 0;

}

void
sync_ping()
{
    struct vi_cseg *m_id, *m_id2;
    int i;

    for (i = 0; i < number; i++) {

	bobnet_recv(packets[1], size, 0);
	bobnet_send(packets[0], size, 0);

	bobnet_sendwait(&m_id, 0);
	if (m_id == NULL)
	    printf("sync_ping: bad sendwait.\n");

	bobnet_recvwait(&m_id2, 0);
	if (m_id2 == NULL)
	    printf("sync_ping: bad recvwait.\n");
    }
}

void
sync_pong()
{
    struct vi_cseg *m_id, *m_id2;
    int i;

    bobnet_recv(packets[0], size, 0);
    for (i = 0; i < number; i++) {
	bobnet_recvwait(&m_id, 0);
	if (m_id == NULL)
	    printf("pong: bad recvwait.\n");

	bobnet_send(packets[1], size, 0);
	bobnet_recv(packets[0], size, 0);

	bobnet_sendwait(&m_id2, 0);
	if (m_id2 == NULL)
	    printf("pong: bad sendwait.\n");
    }
}

void
async_ping()
{
}

void
async_pong()
{
}

/*
 * Internal bobnet functions.
 */
void
bobnet_init(const char *peer_name)
{
    int error, i;
    struct vi_attribs attr;
    struct vi_net_address_ether addr_ether;
    struct vi_net_address *addr = (struct vi_net_address *) &addr_ether;

    error = VipOpenNic(dev, &nic_handle);
    if (error != VIP_SUCCESS) {
	printf("bobnet_init: failed to open nic.\n");
	exit(error);
    }
    error = VipRegisterMem(nic_handle, descs,
			   sizeof(struct vi_cseg) * NDESC * 2, NULL, &desc_mh);
    if (error != VIP_SUCCESS) {
	printf("bobnet_init: failed to register descriptor memory.\n");
	exit(error);
    }

    error = VipRegisterMem(nic_handle, packets, sizeof(packets),
			   NULL, &packet_mh);
    if (error != VIP_SUCCESS) {
	printf("bobnet_init: failed to register packet memory.\n");
	exit (error);
    }

    if (peer_name) {  /* == transmitter */
	/* lookup the destination */
	VipAddressInitializeEther(&addr_ether);
	if (VipGetAddressByName(nic_handle, peer_name, addr) < 0) {
	    fprintf(stderr, "Unknown host \"%s\".\n", peer_name);
	    exit(1);
	}
	addr_ether.port = listen_port;

	/* call up the receiver */
	for (;;) {
	    /* create a vi */
	    memset(&attr, 0, sizeof(attr));
	    attr.ReliabilityLevel = VIP_SERVICE_UNRELIABLE;
	    error = VipCreateVi(nic_handle, &attr, 0, 0, &vi_handle);
	    if (error != VIP_SUCCESS) {
		printf("bobnet_init: failed to create vi.\n");
		exit(error);
	    }

	    error = VipConnectRequest(vi_handle, NULL, addr, 2000, NULL);
	    if (error == VIP_TIMEOUT)   /* timeout is in ms */
		printf("bobnet_init: VipConnectRequest: timeout.\n");
	    else if (error == VIP_SUCCESS) break;
	    else {
		fprintf(stderr, "bobnet_init: VipConnectRequest: error %d\n",
		  error);
		exit(1);
	    }

	    /* destroy it to un-post the receive, and try again (hack) */
	    if ((error = VipDestroyVi(vi_handle)) != VIP_SUCCESS) {
		fprintf(stderr, "bobnet_init: VipDestroyVi: error %d\n", error);
		exit(1);
	    }
	}
    } else {  /* receiver */
	int conn_handle;

	/* wait forever for the transmitter at a well-known port */
	VipAddressInitializeEther(&addr_ether);
	addr_ether.port = listen_port;

	error = VipConnectWait(nic_handle, addr, 0, NULL, NULL,
	  &conn_handle);
	if (error != VIP_SUCCESS) {
	    fprintf(stderr, "bobnet_init: VipConnectWait: error %d\n",
	      error);
	    exit(1);
	}

	/* create a vi */
	memset(&attr, 0, sizeof(attr));
	attr.ReliabilityLevel = VIP_SERVICE_UNRELIABLE;
	error = VipCreateVi(nic_handle, &attr, 0, 0, &vi_handle);
	if (error != VIP_SUCCESS) {
	    printf("bobnet_init: failed to create vi.\n");
	    exit(error);
	}

	/* accept no matter what */
	error = VipConnectAccept(conn_handle, vi_handle);
	if (error != VIP_SUCCESS) {
	    fprintf(stderr, "bobnet_init: VipConnectAccept: error %d\n",
	      error);
	    exit(1);
	}
    }

    free_head = &free_descs[0];
    for (i = 0; i < NDESC; i++) {
	free_descs[i].next = &free_descs[i + 1];
	free_descs[i].desc = &descs[i * 2];
    }
    free_descs[NDESC-1].next = NULL;
}

void
bobnet_finish(void)
{
    int error;

    error = VipDeregisterMem(nic_handle, descs, desc_mh);
    if (error != VIP_SUCCESS) {
	printf("bobnet_finish: failed to deregister descriptors.\n");
	exit (error);
    }

    error = VipDeregisterMem(nic_handle, packets, packet_mh);
    if (error != VIP_SUCCESS) {
	printf("bobnet_finish: failed to deregister packets.\n");
	exit (error);
    }

    error = VipDisconnect(vi_handle);
    if (error != VIP_SUCCESS) {
	fprintf(stderr, "bobnet_finish: VipDisconnect: error %d\n", error);
	exit(1);
    }

    error = VipDestroyVi(vi_handle);
    if (error != VIP_SUCCESS) {
	printf("bobnet_finish: failed to destroy vi.\n");
	exit (error);
    }

    error = VipCloseNic(nic_handle);
    if (error != VIP_SUCCESS) {
	printf("bobnet_finish: failed to close nic.\n");
	exit (error);
    }

}

static struct vi_cseg *
descriptor_next(void)
{
    struct vi_cseg *cseg = free_head->desc;

    if (cseg)
	free_head = free_head->next;
    else
	fprintf(stderr, "descriptor_next: no free descriptors.\n");
    return cseg;
}


int
bobnet_send(void *packet, int len, int dest)
{
    struct vi_cseg *cseg;
    struct vi_dseg *dseg;
    int error;

    cseg = descriptor_next();
    cseg->SegCount = 1;
    cseg->Control = VIP_CONTROL_OP_SENDRECV;
    cseg->Length = len;

    dseg = (struct vi_dseg *)(cseg + 1);
    dseg->Handle = packet_mh;
    dseg->Data.Address = packet;
    dseg->Length = len;

    error = VipPostSend(vi_handle, cseg, desc_mh);
    if (error != VIP_SUCCESS) {
	printf("bobnet_send: error %d while posting descriptor.\n", error);
	return 1;
    }

    return 0;
}

int
bobnet_recv(void *packet, int len, int dest)
{
    struct vi_cseg *cseg;
    struct vi_dseg *dseg;
    int error;

    cseg = descriptor_next();
    cseg->SegCount = 1;
    cseg->Control = VIP_CONTROL_OP_SENDRECV;
    cseg->Length = len;

    dseg = (struct vi_dseg *)(cseg + 1);
    dseg->Handle = packet_mh;
    dseg->Data.Address = packet;
    dseg->Length = len;

    error = VipPostRecv(vi_handle, cseg, desc_mh);
    if (error != VIP_SUCCESS) {
	printf("bobnet_recv: error %d while posting descriptor.\n", error);
	return 1;
    }

    return 0;
}

/*
 * Auto-acking immediate send/recv of 32-bit quantities.
 */
void
bobnet_send_immediate(u_int32_t data, int dest)
{
    struct vi_cseg *cseg, *cseg_recv;
    int err;

    cseg_recv = descriptor_next();
    cseg_recv->SegCount = 1;
    cseg_recv->Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE;
    if ((err=VipPostRecv(vi_handle, cseg_recv, desc_mh)) != VIP_SUCCESS)
	fprintf(stderr, "bobnet_send_immediate: PostRecv error %d", err);

    cseg = descriptor_next();
    cseg->SegCount = 1;
    cseg->Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE;
    cseg->ImmediateData = data;
    if ((err=VipPostSend(vi_handle, cseg, desc_mh)) != VIP_SUCCESS)
	fprintf(stderr, "bobnet_send_immediate: PostSend error %d", err);
    (void) bobnet_sendwait(&cseg, 0);
    (void) bobnet_recvwait(&cseg_recv, 0);
}

void
bobnet_recv_immediate(u_int32_t *data, int dest)
{
    struct vi_cseg *cseg;
    int err;

    cseg = descriptor_next();
    cseg->SegCount = 1;
    cseg->Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE;
    if ((err=VipPostRecv(vi_handle, cseg, desc_mh)) != VIP_SUCCESS)
	fprintf(stderr, "bobnet_recv_immediate: error %d", err);
    (void) bobnet_recvwait(&cseg, 0);
    *data = cseg->ImmediateData;

    cseg = descriptor_next();
    cseg->SegCount = 1;
    cseg->Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE;
    cseg->ImmediateData = 0;  /* ack */
    if ((err=VipPostSend(vi_handle, cseg, desc_mh)) != VIP_SUCCESS)
	fprintf(stderr, "bobnet_send_immediate: PostSend error %d", err);
    (void) bobnet_sendwait(&cseg, 0);
}

int
bobnet_sendwait(struct vi_cseg **desc, int t)
{
    int error, idx;

    error = VipSendWait(vi_handle, t, desc);
    if (error != VIP_SUCCESS) {
	printf("bobnet_sendwait: failed.\n");
	return 1;
    }

    idx = (*desc - descs) / 2;
    if (idx > NDESC) {
	printf("bobnet_sendwait: desc out of range.\n");
	return 1;
    }
    free_descs[idx].next = free_head;
    free_head = &free_descs[idx];

    return 0;
}

int
bobnet_recvwait(struct vi_cseg **desc, int t)
{
    int error, idx;

    error = VipRecvWait(vi_handle, t, desc);
    if (error != VIP_SUCCESS) {
	printf("bobnet_recvwait: failed.\n");
	return 1;
    }

    idx = (*desc - descs) / 2;
    if (idx > NDESC) {
	printf("bobnet_recvwait: desc out of range.\n");
	return 1;
    }
    free_descs[idx].next = free_head;  
    free_head = &free_descs[idx];

    return 0;
}

