/*
 * Support for ethernet hardware devices.
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 *
 * $Id: ether.c,v 1.1 1998/09/22 22:01:14 pw Exp $
 */
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include "ether.h"
#include "libvi.h"

#ifdef __alpha__
  /* libc version definitely broken */
# define NEED_ETHER_ATON
#endif

#if defined(NEED_ETHER_ATON) || defined(NEED_ETHER_COMPAT)

#include <ctype.h>

/* would be static, but declared elsewhere already */
struct ether_addr *
ether_aton_r(const char *asc, struct ether_addr *addr)
{
    size_t cnt;

    for (cnt = 0; cnt < 6; ++cnt) {
	unsigned int number;
	char ch;

	ch = tolower (*asc++);
	if ((ch < '0' || ch > '9') && (ch < 'a' || ch > 'f'))
	    return NULL;
	number = isdigit (ch) ? (ch - '0') : (ch - 'a' + 10);

	ch = tolower (*asc);
	if ((cnt < 5 && ch != ':') || (cnt == 5 && ch != '\0' && !isspace (ch))) {
	    ++asc;
	    if ((ch < '0' || ch > '9') && (ch < 'a' || ch > 'f'))
		return NULL;
	    number <<= 4;
	    number += isdigit (ch) ? (ch - '0') : (ch - 'a' + 10);

	    ch = *asc;
	    if (cnt < 5 && ch != ':')
	    return NULL;
	}

	/* Store result.  */
	addr->ether_addr_octet[cnt] = (unsigned char) number;

	/* Skip ':'.  */
	++asc;
    }

    return addr;
}
#else
#  define ether_aton ether_aton_r
#endif

#ifdef NEED_ETHER_COMPAT
static int
ether_hostton(const char *hostname, struct ether_addr *ea)
{
    FILE *fp;
    char s[2048], *cp, *cq;

    if (!(fp = fopen("/etc/ethers", "r")))
	return -1;
    for (;;) {
	if (!fgets(s, sizeof(s), fp))
	    break;
	for (cp=s; isspace(*cp); cp++) ;
	if (*cp == '#') continue;
	if (!(cq = strchr(cp, ' ')))
	    break;
	for (; isspace(*cq); cq++) ;
	if (!strcasecmp(cq, hostname)) {
	    (void) fclose(fp);
	    return ether_aton_r(cp, ea) ? 0 : -1;
	}
    }
    (void) fclose(fp);
    return -1;
}

#endif

void VipAddressInitializeEther(struct vi_net_address_ether *a)
{
    a->HostAddressLen = VIP_ETHER_ADDRESS_LEN;
    a->DiscriminatorLen = VIP_ETHER_DISCRIMINATOR_LEN;
    memset(&a->ea, 0, sizeof(struct ether_addr));
    a->port = 0;
}

static VIP_RETURN
ether_lookup(const char *name, struct vi_net_address *addr_in)
{
    struct ether_addr ea;
    struct vi_net_address_ether *addr = (struct vi_net_address_ether *)addr_in;

    if (addr->HostAddressLen != VIP_ETHER_ADDRESS_LEN
     || addr->DiscriminatorLen != VIP_ETHER_DISCRIMINATOR_LEN)
       return VIP_INVALID_PARAMETER;

    /* try octets, then host lookup */
    if (!ether_aton_r(name, &ea))
	if (ether_hostton(name, &ea) < 0) {
	    return -1;
	}

    /* addr->HostAddressLen = VIP_ETHER_ADDRESS_LEN; */
    /* addr->DiscriminatorLen = VIP_ETHER_DISCRIMINATOR_LEN; */
    memcpy(&addr->ea, &ea, sizeof(struct ether_addr));
    addr->port = 0;  /* default port before connection is bound? */
    return 0;
}

static struct vi_net_address_ether last_addr;

/*
 * Turn on the filter to listen to any hardware address, at the port
 * specified in local_addr.
 */
static VIP_RETURN
ether_listen(int nic, int vi, const struct vi_net_address *local_addr)
{
    struct vi_bind_req_ether req;

    req.vi = vi;
    /* bind to any source address */
    VipAddressInitializeEther(&req.addr);
    /* but the port to listen on is specified in the net_address */
    if (local_addr)
	req.addr.port = ((struct vi_net_address_ether *)local_addr)->port;
    if (ioctl(nic, VI_BIND, (struct vi_net_address *) &req) < 0) {
	vi_perror("ether_listen: ioctl(%d, VI_BIND)", nic);
	return VIP_INVALID_PARAMETER;
    }
    return VIP_SUCCESS;
}

/*
 * When in listen mode, will give back the last received connection attempt.
 */
static VIP_RETURN
ether_getpeerinfo(int nic, int vi, const struct vi_net_address **remote_addr)
{
    struct vi_getpeer_req_ether req;

    req.vi = vi;
    VipAddressInitializeEther((struct vi_net_address_ether *) &req.addr);
    if (ioctl(nic, VI_GETPEER, &req) < 0) {
	vi_perror("ether.getpeerinfo: ioctl(%d, VI_GETPEER)", nic);
	return VIP_INVALID_PARAMETER;
    }
    memcpy(&last_addr, &req.addr, sizeof(struct vi_net_address_ether));
    *remote_addr = (struct vi_net_address *) &last_addr;
    return VIP_SUCCESS;
}

/*
 * Filter only to accept from remote_addr.
 */
static VIP_RETURN
ether_bind(int nic, int vi, const struct vi_net_address *remote_addr)
{
    struct vi_bind_req_ether req;

    req.vi = vi;
    memcpy(&req.addr, remote_addr, sizeof(struct vi_net_address_ether));
    if (ioctl(nic, VI_BIND, (struct vi_net_address *) &req) < 0) {
	vi_perror("ether_bind: ioctl(%d, VI_BIND)", nic);
	return VIP_INVALID_PARAMETER;
    }
    return VIP_SUCCESS;
}

static int
ether_address_length(void)
{
    return sizeof(struct vi_net_address_ether);
}

struct hardware_desc hardware_ether = {
           type:	HW_ETHER,
         listen:	ether_listen,
    getpeerinfo:	ether_getpeerinfo,
           bind:	ether_bind,
         lookup:	ether_lookup,
 address_length:	ether_address_length,
};

