/*
 * Main VI implementation library.
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include "extern.h"  /* declarations of the functions defined here */
#include "libvi.h"

#define MIN(a,b) ((a)<(b)?(a):(b))

/* known hardware list */
struct hardware_desc hardware_none = { HW_NONE };

#ifdef CONFIG_HARDWARE_ETHER
extern struct hardware_desc hardware_ether;
#endif

struct hardware_desc *hardware[] = {
    &hardware_none,
#ifdef CONFIG_HARDWARE_ETHER
    &hardware_ether,
#endif
};
#define NUM_HARDWARE (sizeof(hardware)/sizeof(hardware[0]))

/*
 * One entry for each open nic.
 */
struct nic_info {
    struct nic_info *next, **pprev;
    int nic;
    hw_type hw;
};
static struct nic_info *nics = 0;

/*
 * User space array to keep track of work queues/etc..
 */
struct viq {
    struct vi_cseg *head;
    struct vi_cseg *tail;
    int compq;
};

/*
 * One entry for each open vi.
 */
struct vi_info {
    struct vi_info *next, **pprev;
    int vi;
    int nic;
    struct viq sendq, recvq;
};
static struct vi_info *vis = 0;

/*
 * Error handling.
 */
static char *vip_errstr[] = {
    "VIP_SUCCESS",
    "VIP_NOT_DONE",
    "VIP_INVALID_PARAMETER",
    "VIP_ERROR_RESOURCE",
    "VIP_TIMEOUT",
    "VIP_REJECT",
    "VIP_INVALID_RELIABILITY_LEVEL",
    "VIP_INVALID_MTU",
    "VIP_INVALID_QOS",
    "VIP_INVALID_PTAG",
    "VIP_INVALID_RDMAREAD",
};
#define NUM_VIP_ERRSTR (sizeof(vip_errstr)/sizeof(vip_errstr[0]))

/*
 * Locally generated VIPL errors.
 */
VIP_RETURN
vi_verror(VIP_RETURN err, const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    if (err < 0 || err >= NUM_VIP_ERRSTR)
	fprintf(stderr, ": Unknown error %d.\n", err);
    else
	fprintf(stderr, ": %s.\n", vip_errstr[err]);
    return err;
}

void
vi_warn(const char *fmt, ...)
{
    va_list ap;

    fprintf(stderr, "Warning: ");
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
}

void
vi_error(const char *fmt, ...)
{
    va_list ap;

    fprintf(stderr, "Error: ");
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}

/*
 * Libc errors from errno.
 */
VIP_RETURN
vi_perror(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, ": %s.\n", strerror(errno));
    return errno;
}

/*
 * Safe malloc
 */
void *
Malloc(unsigned int n)
{
    void *x;

    if (!(x = malloc(n))) {
	vi_perror("Malloc: %d bytes", n);
	exit(1);
	return NULL;
    }
    return x;
}

/*
 * Local utility funcs
 */
static struct nic_info *
vi_get_nic_info(int nic)
{
    struct nic_info *ni;

    for (ni=nics; ni; ni=ni->next)
	if (ni->nic == nic)
	    return ni;
    return NULL;
}

static struct vi_info *
vi_get_vi_info(int vi)
{
    struct vi_info *vii;

    for (vii=vis; vii; vii=vii->next)
	if (vii->vi == vi)
	    return vii;
    return NULL;
}

static int
vi_nic_from_vi(int vi)
{
    struct vi_info *vii;

    if (!(vii = vi_get_vi_info(vi)))
	vi_error("vi_nic_from_vi: unknown vi %d", vi);
    return vii->nic;
}

static hw_type
vi_hw_from_nic(int nic)
{
    struct nic_info *ni;

    if (!(ni = vi_get_nic_info(nic)))
	vi_error("vi_nic_from_nic: unknown nic %d", nic);
    return ni->hw;
}

/*
 * Hardware connection:  open, close.
 */
VIP_RETURN
VipOpenNic(const char *dev_name, int *nic_handle)
{
    if ((*nic_handle = open(dev_name, O_RDWR, 0)) < 0) {
	(void) vi_perror("VipOpenNic: open(\"%s\", RDWR)", dev_name);
	return VIP_INVALID_PARAMETER;
    } else {
	struct nic_info *ni;

	ni = Malloc(sizeof(struct nic_info));
	ni->nic = *nic_handle;
	ni->hw = HW_ETHER;  /* kludge, should call dev */
	ni->pprev = NULL;
	ni->next = nics;
	if (ni->next)
	    ni->next->pprev = &ni->next;
	nics = ni;
	return VIP_SUCCESS;
    }
}

VIP_RETURN
VipCloseNic(int nic_handle)
{
    if (close(nic_handle) < 0) {
	(void) vi_perror("VipCloseNic: close(%d)", nic_handle);
	return VIP_INVALID_PARAMETER;
    } else {
	struct nic_info *ni;

	if (!(ni = vi_get_nic_info(nic_handle)))
	    vi_error("VipCloseNic: no nic_info for handle %d", nic_handle);
	if (ni->next)
	    ni->next->pprev = ni->pprev;
	if (ni->pprev)
	    *ni->pprev = ni->next;
	free(ni);
	return VIP_SUCCESS;
    }
}

/*
 * Endpoint creation/destruction.
 */
VIP_RETURN
VipCreateVi(int nic_handle, const struct vi_attribs *ViAttribs,
	    int SendCQHandle, int RecvCQHandle, int *vi_handle)
{
    struct vi_create_req req;
    struct vi_info *vii;

    if (ViAttribs != NULL)
	req.ViAttribs = *ViAttribs;
    req.SendCQHandle = SendCQHandle;
    req.RecvCQHandle = RecvCQHandle;
    req.hw = vi_hw_from_nic(nic_handle);

    if (ioctl(nic_handle, VI_CREATE, &req) < 0) {
	(void) vi_perror("VipCreateVi: ioctl(%d, VI_CREATE)", nic_handle);
	return VIP_INVALID_PARAMETER;
    }

    *vi_handle = req.ViHandle;

    vii = Malloc(sizeof(struct vi_info));
    vii->vi = *vi_handle;
    vii->nic = nic_handle;
    vii->pprev = NULL;
    vii->next = vis;
    vii->sendq.head = vii->sendq.tail = NULL;
    vii->sendq.compq = -1;
    vii->recvq.head = vii->recvq.tail = NULL;
    vii->recvq.compq = -1;
    if (vii->next)
	vii->next->pprev = &vii->next;
    vis = vii;
    return VIP_SUCCESS;
}

VIP_RETURN
VipDestroyVi(int vi_handle)
{
    int nic;
    struct vi_info *vii;

    nic = vi_nic_from_vi(vi_handle);
    if (ioctl(nic, VI_DESTROY, &vi_handle) < 0) {
	(void) vi_perror("VipDestroyVi(%d): ioctl VI_DESTROY", vi_handle);
	return VIP_INVALID_PARAMETER;
    }

    if (!(vii = vi_get_vi_info(vi_handle)))
	vi_error("VipDestroyVi: no vi_info for handle %d", vi_handle);
    if (vii->next)
	vii->next->pprev = vii->pprev;
    if (vii->pprev)
	*vii->pprev = vii->next;
    free(vii);
    return VIP_SUCCESS;
}

static int connect_handle = 0;
static const struct vi_net_address *connect_addr;
static int connect_vi;  /* the vi that was calling */

/*
 * Connection Management.  Very ether-specific now.  Need to generalize
 * and move stuff into ether.c later.
 */
VIP_RETURN
VipConnectWait(int nic_handle,
  const struct vi_net_address *local_addr, unsigned long timeout,
  struct vi_net_address *remote_addr, struct vi_attribs *remote_attr,
  int *conn_handle_return)
{
    int err, mem_handle, vi_handle;
    struct vi_cseg cs, *csret;
    struct hardware_desc *hw = hardware[vi_hw_from_nic(nic_handle)];

    /*
     * Listen on well-known port 0, for any host.
     */
    if ((err = VipCreateVi(nic_handle, NULL, 0, 0, &vi_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectWait: VipCreateVi");

    if ((err = (hw->listen)(nic_handle, vi_handle, local_addr)) != VIP_SUCCESS)
        return vi_verror(err, "VipConnectWait: hardware.listen");

    if ((err = VipRegisterMem(nic_handle, &cs, sizeof(struct vi_cseg), NULL,
      &mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectWait: VipRegisterMem");

    cs.SegCount = 0;
    cs.Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE;
    if ((err = VipPostRecv(vi_handle, &cs, mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectWait: VipPostRecv");

    if ((err = VipRecvWait(vi_handle, timeout, &csret)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectWait: VipRecvWait");

    if ((err = VipDeregisterMem(nic_handle, &cs, mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectWait: VipDeregisterMem");

    /* who is on the other side */
    if ((err = (hw->getpeerinfo)(nic_handle, vi_handle, &connect_addr))
      != VIP_SUCCESS)
        return vi_verror(err, "VipConnectWait: hardware.getpeerinfo");

    if (remote_addr) {
	int len = MIN(
	   remote_addr->HostAddressLen +  remote_addr->DiscriminatorLen,
	  connect_addr->HostAddressLen + connect_addr->DiscriminatorLen);
	memcpy(remote_addr, connect_addr, len);
    }

    if (remote_attr)
	vi_warn("VipConnectWait: can't copy remote_attr");

    if ((err = VipDestroyVi(vi_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectWait: ViDestroyVi");

    *conn_handle_return = ++connect_handle;
    return VIP_SUCCESS;
}

VIP_RETURN
VipConnectAccept(int conn_handle, int vi_handle)
{
    int err, mem_handle;
    struct vi_cseg cs, *csret;
    int nic_handle = vi_nic_from_vi(vi_handle);
    struct hardware_desc *hw = hardware[vi_hw_from_nic(nic_handle)];

    if (conn_handle != connect_handle)
	return VIP_INVALID_PARAMETER;
    /*
     * Reset to listen to whatever was specified in the offered connection.
     */
    if ((err = (hw->bind)(nic_handle, vi_handle, connect_addr)) != VIP_SUCCESS)
        return vi_verror(err, "VipConnectAccept: hardware.bind");

    if ((err = VipRegisterMem(nic_handle, &cs, sizeof(struct vi_cseg), NULL,
      &mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectAccept: VipRegisterMem");

    cs.SegCount = 0;
    cs.Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE
               | VIP_CONTROL_CONNECT;
    cs.ImmediateData = VIP_SUCCESS;
    if ((err = VipPostSend(vi_handle, &cs, mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectAccept: VipPostRecv");

    if ((err = VipSendWait(vi_handle, 0, &csret)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectAccept: VipSendWait");

    if ((err = VipDeregisterMem(nic_handle, &cs, mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectAccept: VipDeregisterMem");

    return VIP_SUCCESS;
}

VIP_RETURN
VipConnectReject(int conn_handle)
{
    int err, mem_handle;
    struct vi_cseg cs, *csret;
    int vi_handle = connect_vi;
    int nic_handle = vi_nic_from_vi(vi_handle);
    struct hardware_desc *hw = hardware[vi_hw_from_nic(nic_handle)];

    printf("VipConnectReject\n");

    if (conn_handle != connect_handle)
	return VIP_INVALID_PARAMETER;
    /*
     * Reset to talk to whatever was specified in the offered connection,
     * but just so we can reply in the negative.
     */
    if ((err = (hw->bind)(nic_handle, vi_handle, connect_addr)) != VIP_SUCCESS)
        return vi_verror(err, "VipConnectReject: hardware.bind");

    if ((err = VipRegisterMem(nic_handle, &cs, sizeof(struct vi_cseg), NULL,
      &mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectReject: VipRegisterMem");

    cs.SegCount = 0;
    cs.Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE;
    cs.ImmediateData = VIP_REJECTED;
    if ((err = VipPostSend(vi_handle, &cs, mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectReject: VipPostRecv");

    if ((err = VipSendWait(vi_handle, 0, &csret)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectReject: VipSendWait");

    if ((err = VipDeregisterMem(nic_handle, &cs, mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectReject: VipDeregisterMem");

    if ((err = (hw->listen)(nic_handle, vi_handle, NULL)) != VIP_SUCCESS)
        return vi_verror(err, "VipConnectReject: hardware.listen");

    return VIP_SUCCESS;
}

/*
 * Returns actual remote_addr, since the port may be different than the
 * calling port.
 */
VIP_RETURN
VipConnectRequest(int vi_handle,
  const struct vi_net_address *local_addr,
  struct vi_net_address *remote_addr, unsigned long timeout,
  struct vi_attribs *remote_attr)
{
    int err, mem_handle;
    struct vi_cseg cs[2], *csret;
    int nic_handle = vi_nic_from_vi(vi_handle);
    struct hardware_desc *hw = hardware[vi_hw_from_nic(nic_handle)];

    if ((err = (hw->bind)(nic_handle, vi_handle, remote_addr)) != VIP_SUCCESS)
        return vi_verror(err, "VipConnectRequest: hardware.bind");

    /*
     * Now do a "ping", hoping somebody is listening.
     */
    if ((err = VipRegisterMem(nic_handle, cs, 2*sizeof(struct vi_cseg), NULL,
      &mem_handle)) != VIP_SUCCESS)
	return vi_verror(err, "VipConnectRequest: VipRegisterMem");

    memset(cs, 0, 2 * sizeof(struct vi_cseg));
    cs[0].SegCount = 0;
    cs[0].Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE;
    if ((err = VipPostRecv(vi_handle, &cs[0], mem_handle)) != VIP_SUCCESS) {
	(void) vi_verror(err, "VipConnectRequest: VipPostRecv");
	goto out;
    }

    cs[1].SegCount = 0;
    cs[1].Control = VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE
                  | VIP_CONTROL_CONNECT;
    cs[1].ImmediateData = 0;
    if ((err = VipPostSend(vi_handle, &cs[1], mem_handle)) != VIP_SUCCESS) {
	(void) vi_verror(err, "VipConnectRequest: VipPostSend");
	goto out;
    }

    if ((err = VipSendWait(vi_handle, timeout, &csret)) != VIP_SUCCESS) {
	(void) vi_verror(err, "VipConnectRequest: VipSendWait");
	goto out;
    }
    if (csret != &cs[1])
	printf("send expecting csret = %p would be &cs[1] = %p\n",
	  csret, &cs[1]);

    if ((err = VipRecvWait(vi_handle, timeout, &csret)) != VIP_SUCCESS) {
	(void) vi_verror(err, "VipConnectRequest: VipRecvWait");
	goto out;
    }
    if (csret != &cs[0])
	printf("recv expecting csret = %p would be &cs[0] = %p\n",
	  csret, &cs[0]);

    /* rebind to the new port number returned */
    if (cs[0].ImmediateData == VIP_SUCCESS) {
	const struct vi_net_address *actual_addr;
	if ((err = (hw->getpeerinfo)(nic_handle, vi_handle, &actual_addr))
	  != VIP_SUCCESS)
	    return vi_verror(err, "VipConnectRequest: hardware.getpeerinfo");
	memcpy(remote_addr, actual_addr, (hw->address_length)());
	if ((err = (hw->bind)(nic_handle, vi_handle, remote_addr))
	  != VIP_SUCCESS)
	    return vi_verror(err, "VipConnectRequest: hardware.bind #2");
	err = VIP_SUCCESS;
    } else if (cs[0].ImmediateData == VIP_REJECTED) {
	err = VIP_REJECTED;
    } else
	(void) vi_verror(err, "VipConnectRequest: connection status");

  out:
    {
    int err2;
    if ((err2=VipDeregisterMem(nic_handle, cs, mem_handle)) != VIP_SUCCESS)
	return vi_verror(err2, "VipConnectRequest: VipDeregisterMem");
    }

    return err;

}

VIP_RETURN
VipDisconnect(int vi_handle)
{
    int err;
    int nic_handle = vi_nic_from_vi(vi_handle);
    struct hardware_desc *hw = hardware[vi_hw_from_nic(nic_handle)];

    if ((err = (hw->listen)(nic_handle, vi_handle, NULL)) != VIP_SUCCESS)
        return vi_verror(err, "VipDisconnect: hardware.listen");
    return VIP_SUCCESS;
}

/*
 * Memory registration
 */
VIP_RETURN
VipRegisterMem(int NicHandle, void *VirtualAddress, unsigned long Length,
	       const struct vi_mem_attr *MemAttribs, int *MemoryHandle)
{
    struct vi_memreg_req req;

    req.nic_handle = NicHandle;
    req.addr = (caddr_t) VirtualAddress;
    req.len = Length;
    if (MemAttribs != NULL)
	req.mem_attribs = *MemAttribs;

    if (ioctl(NicHandle, VI_MEM_REG, &req) < 0) {
	(void) vi_perror("VipRegisterMem: ioctl(%d, VI_MEM_REG)", NicHandle);
	return VIP_INVALID_PARAMETER;
    }
 
    *MemoryHandle = req.mem_handle;
    return VIP_SUCCESS;
}
 
VIP_RETURN 
VipDeregisterMem(int NicHandle, const void *VirtualAddress, int MemoryHandle)
{
    struct vi_memreg_req req;
    
    req.nic_handle = NicHandle;
    req.addr = (caddr_t) VirtualAddress;
    req.mem_handle = MemoryHandle;
    
    if (ioctl(NicHandle, VI_MEM_DEREG, &req) < 0) {
	(void) vi_perror("VipDeregisterMem: ioctl(%d, VI_MEM_DEREG)",
	  NicHandle);
	return VIP_INVALID_PARAMETER;
    }
    
    return VIP_SUCCESS;
}   

/*
 * Send/receive posting
 */
static VIP_RETURN
generic_post(struct viq *q, struct vi_cseg *cs, int vi_handle, int mem_handle,
  int send)
{
    struct vi_trap_args args;

    cs->Status = 0;
    cs->Next.Address = 0;
    if (q->head == NULL) {
	q->head = cs;
	q->tail = cs;
    } else {
	struct vi_cseg *last = q->tail;
	last->NextHandle = mem_handle;
	last->Next.Address = cs;
	q->tail = cs;
	if (last->Status == 0) {
	    printf("VipPost%s: last status 0, skipping trap.\n",
	      send ? "Send" : "Recv");
	    return VIP_SUCCESS;
	}
    }

    args.cs = cs;
    args.mh_idx = mem_handle;
    args.vi_idx = vi_handle;
    args.send = send;
    VI_TRAP(&args);

    return VIP_SUCCESS;
}

VIP_RETURN
VipPostSend(int vi_handle, struct vi_cseg *cs, int mem_handle)
{
    struct vi_info *vii;

    if (!(vii=vi_get_vi_info(vi_handle)))
	vi_error("VipPostSend: unknown vi %d", vi_handle);
    return generic_post(&vii->sendq, cs, vi_handle, mem_handle, 1);

}

VIP_RETURN
VipPostRecv(int vi_handle, struct vi_cseg *cs, int mem_handle)
{
    struct vi_info *vii;

    if (!(vii=vi_get_vi_info(vi_handle)))
	vi_error("VipPostRecv: unknown vi %d", vi_handle);
    return generic_post(&vii->recvq, cs, vi_handle, mem_handle, 0);
}

/*
 * Done functions
 */
static VIP_RETURN
generic_done(struct viq *q, struct vi_cseg **cs)
{
    if (q->head != NULL && q->head->Status != 0) {
	*cs = q->head;
	q->head = q->head->Next.Address;
	return VIP_SUCCESS;
    }
    *cs = NULL;
    return VIP_NOT_DONE;
}

VIP_RETURN
VipSendDone(int vi_handle, struct vi_cseg **cs)
{
    struct vi_info *vii;

    if (!(vii=vi_get_vi_info(vi_handle)))
	vi_error("VipSendDone: unknown vi %d", vi_handle);
    return generic_done(&vii->sendq, cs);
}

VIP_RETURN
VipRecvDone(int vi_handle, struct vi_cseg **cs)
{
    struct vi_info *vii;

    if (!(vii=vi_get_vi_info(vi_handle)))
	vi_error("VipRecvDone: unknown vi %d", vi_handle);
    return generic_done(&vii->recvq, cs);
}

/*
 * Waiting functions
 */
static VIP_RETURN
generic_wait(struct viq *q, struct vi_cseg **cs, int timeout)
{
    if (q->head == NULL) {
	*cs = NULL;
	return VIP_INVALID_PARAMETER;
    }

    if (timeout == 0) {
	while (q->head->Status == 0)
	    ;
    } else {
	for (; timeout; timeout--) {
	    if (q->head->Status != 0)
		goto done;
	    usleep(1000);
	}
	*cs = NULL;
	return VIP_TIMEOUT;
    }

done:
    *cs = q->head;
    q->head = q->head->Next.Address;
    return VIP_SUCCESS;
}

VIP_RETURN
VipSendWait(int vi_handle, unsigned long timeout, struct vi_cseg **cs)
{
    struct vi_info *vii;

    if (!(vii=vi_get_vi_info(vi_handle)))
	vi_error("VipSendWait: unknown vi %d", vi_handle);
    return generic_wait(&vii->sendq, cs, timeout);
}

VIP_RETURN
VipRecvWait(int vi_handle, unsigned long timeout, struct vi_cseg **cs)
{
    struct vi_info *vii;

    if (!(vii=vi_get_vi_info(vi_handle)))
	vi_error("VipRecvWait: unknown vi %d", vi_handle);
    return generic_wait(&vii->recvq, cs, timeout);
}

/*
 * Other handy extensions.
 */
VIP_RETURN
VipGetAddressByName(int nic_handle, const char *s, struct vi_net_address *addr)
{
    return (hardware[vi_hw_from_nic(nic_handle)]->lookup)(s, addr);
}
