/*
 * Generic ethernet driver using skb interface.
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <linux/config.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <asm/uaccess.h>
#include "vivar.h"
#include "ether.h"

#define M_NONCONTIG	0x0010	/* m_ext may not contain contiguous data */
#define ETHERTYPE_VIA	0x8000	/* VIA Ethernet frame type */

#if 0
static void
hexdump(void *vwhere, int n)
{
    int i;
    unsigned char *where = (char *)vwhere;

    while (n > 0) {
	printk("%p ", where);
	for (i=0; i < ( (n>16)?16:n); i++)
	    printk(" %02x", where[i]);
	printk("\n");
	n -= 16;
	where += 16;
    }
}
#endif

#define MIN(a,b) ((a) < (b) ? (a) : (b))

int
ether_transmit(struct vi_instance *v, struct vi_cseg *cs)
{
    struct sk_buff *skb;
    struct vi_hdr *vh;
    struct device *dev = v->ifp;
    int len, err, notdone = 0, offset = 0;
    caddr_t data;

    /* see net/packet/af_packet.c */
    
    if (cs->Control & VIP_CONTROL_IMMEDIATE) {
	len = 4;
	data = 0;  /* shut up gcc */
    } else {
	struct vi_dseg *ds = (struct vi_dseg *)(cs + 1);
	if (ds->Length > v->mtu) {
	    len = MIN(ds->Length - v->send_off, v->mtu);
	    offset = v->send_off;
	    v->send_off += len;
	    /* printk("offset %d, len %d, new send_off %d, ds->Length %d\n", */
	      /* offset, len, v->send_off, ds->Length); */
	    data = (caddr_t) ds->Data.Address + v->send_off;
	    if (v->send_off < ds->Length)
		notdone = 1;
	    else
		v->send_off = 0;
	} else {
	    len = ds->Length;
	    data = ds->Data.Address;
	}
    }

    if (!(skb = alloc_skb(
      sizeof(struct ethhdr) + sizeof(struct vi_hdr) + len, GFP_KERNEL))) {
	err = -ENOBUFS;
	goto out_unlock;
    }

    memcpy(skb_put(skb, sizeof(struct ethhdr)), v->header_template,
      sizeof(struct ethhdr));
    vh = (struct vi_hdr *) skb_put(skb, sizeof(struct vi_hdr));
    vh->svi = v->svi;
    vh->dvi = v->dvi;
    vh->len = len;
    vh->off = offset;
    if (cs->Control & VIP_CONTROL_CONNECT)
	vh->flags = VI_CONNECT;
    if (cs->Control & VIP_CONTROL_IMMEDIATE)
	*(u_int32_t *)skb_put(skb,vh->len) = cs->ImmediateData;
    else  /* use vi_utok to get the data in from user space.  memcpy_iovec */
	memcpy(skb_put(skb,vh->len), data, vh->len);

    /* printk("ether_transmit: skb to hard_start_xmit\n");
     * hexdump(skb->data, skb->len);
     */
    dev->hard_start_xmit(skb, dev);

    /* printk("return %d\n", notdone); */
    return notdone;

  out_unlock:
    dev_unlock_list();
    return err;
}

static struct device *hwdev;
static const char *dev = "eth0";
static struct packet_type pt;
static void (*input_func)(struct sk_buff *skb) = 0;

/*
 * Until attached for real, no-op on the packet tracing.
 */
static int
pack_nofunc(struct sk_buff *skb, struct device *dev, struct packet_type *pt)
{
    if (input_func)
	(input_func)(skb);
    return 0;
}

/*
 * See if we're interested in fielding action on the dev.
 */
static int
ether_attach(struct device *dev, void (*func)(struct sk_buff *skb))
{
    if (hwdev->ifindex == dev->ifindex) {
	MOD_INC_USE_COUNT;
	input_func = func;
	return 0;
    } else
	return -ENODEV;
}

static void
ether_detach(void)
{
    MOD_DEC_USE_COUNT;
}

/*
 * Fill in the private fields on a new vi instance.
 */
static void
ether_create(struct vi_instance *v, struct device *dev)
{
    struct ethhdr *eh;

    eh = v->header_template = kmalloc(sizeof(struct ethhdr), GFP_KERNEL);
#ifdef __FreeBSD__
    eh->ether_type = htons(ETHERTYPE_VIA);
    bcopy(((struct arpcom *)sc->ifp)->ac_enaddr, eh->ether_shost, 6);
#endif
#ifdef __linux__
    eh->h_proto = htons(ETHERTYPE_VIA);
    memcpy(eh->h_source, dev->dev_addr, ETH_ALEN);
#endif
    v->peer_addr = kmalloc(sizeof(struct vi_net_address_ether), GFP_KERNEL);
}

static void
ether_destroy(struct vi_instance *v)
{
    kfree(v->header_template);
    kfree(v->peer_addr);
}

static int
ether_bind(struct vi_instance *v, struct vi_bind_req *req,
  struct vi_bind_req *userreq)
{
    struct ethhdr *eh;
    struct vi_net_address_ether addr;
    int i;

    if (req->addr.HostAddressLen != VIP_ETHER_ADDRESS_LEN
     || req->addr.DiscriminatorLen != VIP_ETHER_DISCRIMINATOR_LEN)
       return -EINVAL;
    copy_from_user(&addr, &userreq->addr, sizeof(struct vi_net_address_ether));
    eh = v->header_template;
#ifdef __FreeBSD__
    bcopy(req->ether_dhost, eh->ether_dhost, 6);
#endif
#ifdef __linux__
    memcpy(eh->h_dest, &addr.ea, ETH_ALEN);
#endif
    for (i=0; i<ETH_ALEN; i++)
	if (addr.ea.ether_addr_octet[i] != '\0') break;
    if (i == ETH_ALEN)
	v->state = VI_CONNECTING;  /* all 0s means listening for somebody */
    else
	v->state = VI_CONNECTED;  /* found a peer, exact number */
    v->dvi = addr.port;
    return 0;
}

static int
ether_getpeer(struct vi_instance *v, struct vi_getpeer_req *req,
  struct vi_getpeer_req *userreq)
{
    if (req->addr.HostAddressLen != VIP_ETHER_ADDRESS_LEN
     || req->addr.DiscriminatorLen != VIP_ETHER_DISCRIMINATOR_LEN)
       return -EINVAL;
    copy_to_user(&userreq->addr, v->peer_addr,
      sizeof(struct vi_net_address_ether));
    return 0;
}

void
VipAddressInitializeEther(struct vi_net_address_ether *a)
{
    a->HostAddressLen = VIP_ETHER_ADDRESS_LEN;
    a->DiscriminatorLen = VIP_ETHER_DISCRIMINATOR_LEN;
    memset(&a->ea, 0, sizeof(struct ether_addr));
    a->port = 0;
}

/*
 * Grab the peer parameters from the skb.
 */
static void
ether_store_peer(struct vi_instance *v, struct sk_buff *skb)
{
    struct vi_net_address_ether *na;

    na = (struct vi_net_address_ether *) v->peer_addr;
    VipAddressInitializeEther(na);
    memcpy(&na->ea, skb->mac.ethernet->h_source, ETH_ALEN);
    na->port = ((struct vi_hdr *)skb->data)->svi;
}

static struct hardware_desc hardware_ether = {
     attach:	ether_attach,
     detach:	ether_detach,
     create:	ether_create,
    destroy:	ether_destroy,
       bind:	ether_bind,
    getpeer:	ether_getpeer,
 store_peer:	ether_store_peer,
   transmit:	ether_transmit,
};

MODULE_AUTHOR("Pete Wyckoff <wyckoff@ca.sandia.gov>");
MODULE_DESCRIPTION("Bobnet VI interface, ethernet, through sk_buff");
MODULE_PARM(dev, "s");

/*
 * insert/remove module
 */
int
init_module(void)
{
    int err;

    hwdev = dev_get(dev);
    if (!hwdev) {
	printk("Unknown device \"%s\".\n", dev);
	return -ENODEV;
    }
    pt.type = htons(ETHERTYPE_VIA);
    pt.dev = hwdev;
    pt.func = pack_nofunc;
    pt.data = 0;
    pt.next = 0;
    dev_add_pack(&pt);

    err = register_hardware(HW_ETHER, &hardware_ether);
    if (err < 0)
	return err;
    return 0;
}

void
cleanup_module(void)
{
    (void) unregister_hardware(HW_ETHER);
    dev_remove_pack(&pt);
}

/* vim: set tags+=/usr/src/linux/tags: */
