/*
 * Stubs to convert MPI calls to mplite
 * (Use mpi include path, but don't link to mpi library)
 *    Message types are not used
 *    Only MPI_COMM_WORLD is supported
 */
#include "./mpi.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "mpfuncs.h"

int nbytes;

static int bytesper(MPI_Datatype datatype)
{
   if( datatype == MPI_BYTE ) return 1;
   else if( datatype == MPI_CHAR ) return sizeof(char);
   else if( datatype == MPI_SHORT ) return sizeof(short);
   else if( datatype == MPI_INT ) return sizeof(int);
   else if( datatype == MPI_LONG ) return sizeof(long);
   else if( datatype == MPI_FLOAT ) return sizeof(float);
   else if( datatype == MPI_DOUBLE ) return sizeof(double);
   else {
      fprintf(stderr,"ERROR - Unsupported MPI_Datatype in stubs.c\n");
      exit(-1);
      return 0;
   }
}

static void Check_Communicator(MPI_Comm comm)
{
   if( comm != MPI_COMM_WORLD ) {
      fprintf(stderr,"ERROR - Only MPI_COMM_WORLD is supported\n");
      exit(-1);
   }
}

int MPI_Init(int *argc, char ***argv)
{ MP_Init(argc, argv); return MPI_SUCCESS; }

int MPI_Finalize()
{ MP_Finalize(); return MPI_SUCCESS; }


int MPI_Comm_rank(MPI_Comm comm, int *rank)
{
   Check_Communicator(comm);
   MP_Myproc(rank);
   return MPI_SUCCESS;
}

int MPI_Comm_size(MPI_Comm comm, int *size)
{
   Check_Communicator(comm);
   MP_Nprocs(size);
   return MPI_SUCCESS;
}


int MPI_Send(void *buf, int count, MPI_Datatype datatype, 
             int destination, int mtype, MPI_Comm comm)
{
   Check_Communicator(comm);
   nbytes = count * bytesper(datatype);
   MP_Send(buf, nbytes, destination);   /* Hope mtype is not needed */
   return MPI_SUCCESS;
}

int MPI_Isend(void* buf, int count, MPI_Datatype datatype, 
             int destination, int mtype, MPI_Comm comm, MPI_Request* msg_id)
{
   Check_Communicator(comm);
   nbytes = count * bytesper(datatype);
   MP_ASend(buf, nbytes, destination, msg_id); /* Hope mtype is not needed */
   return MPI_SUCCESS;
}

int MPI_Recv(void *buf, int count, MPI_Datatype datatype, 
             int source, int mtype, MPI_Comm comm, MPI_Status* status)
{ 
   Check_Communicator(comm);
   nbytes = count * bytesper(datatype);
   MP_Recv(buf, nbytes, source);   /* Hope mtype is not needed */
   return MPI_SUCCESS;
}

int MPI_Irecv(void *buf, int count, MPI_Datatype datatype, 
             int source, int mtype, MPI_Comm comm, MPI_Request* msg_id)
{ 
   Check_Communicator(comm);
   nbytes = count * bytesper(datatype);
   MP_ARecv(buf, nbytes, source, msg_id);   /* Hope mtype is not needed */
   return MPI_SUCCESS;
}

int MPI_Wait(MPI_Request* msg_id, MPI_Status* status)
{ MP_Wait(msg_id); return MPI_SUCCESS; }


int MPI_Barrier(MPI_Comm comm)
{
   Check_Communicator(comm);
   MP_Sync();
   return MPI_SUCCESS;
}

double MPI_Wtime()
{
/*  extern double MP_Time();*/
/*printf("In MPI_Wtime() time = %f\n",MP_Time());*/
   return MP_Time();
}


/* Global functions */

int MPI_Allreduce(void *buf1, void *buf2, int count, MPI_Datatype datatype,
               MPI_Op op, MPI_Comm comm)
{
   int opcode;

   Check_Communicator(comm);
   if( op == MPI_SUM ) opcode = 0;
   else if( op == MPI_MAX ) opcode = 2;
   else if( op == MPI_MIN ) opcode = 4;
   else {
      fprintf(stderr,"ERROR in MPI_Allreduce(), unsupported MPI_Op\n");
      exit(-1);
   }

   if( datatype == MPI_INT && sizeof(int) == 4) opcode += 0;
   else if( datatype == MPI_LONG && sizeof(long) == 4) opcode += 0;
   else if( datatype == MPI_DOUBLE ) opcode += 1;
   else {
      fprintf(stderr,"ERROR in MPI_Allreduce(), unsupported MPI_Datatype\n");
      exit(-1);
   }

   memcpy(buf2,buf1,count*bytesper(datatype));
/*   MP_Global(buf2, count, opcode);*/

   return MPI_SUCCESS;
}

