/* MP_Lite by Dave Turner - Ames Lab - Sept of 1997 +++ */

/* This file is for the TCP version which:
 *   - Uses a .mplite.config file to input nprocs and the hostnames
 *   - Uses asynchronous sockets for the basic MP_Send() and MP_Recv() commands
 *   - Uses blocking sockets for MP_Sync()
 *   - Copies to a send buffer if necessary to avoid lockup conditions
 *
 *   IRIX 6.2 Bug note: SIGIO is not generated when a TCP buffer gets emptied.
 *      Setting the buffer size to 4 MB should prevent this from occuring.
 *      Be aware that a lockup may occur messages larger than this are sent.
 *      If you're pushing this limit, simply increase the buffer_size variable.
 */

#include "mplite.h"   /* Globals myproc, nprocs, mylog, hostname, logfile */
#include "mpfuncs.h"

#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#ifdef __FreeBSD__
#  include <sys/syslimits.h>  /* OPEN_MAX on bsd */
#endif
/* not sure if this is what the authors intended */
#ifdef __linux__
#  define O_ASYNC FASYNC
#endif

static void do_send(int node);
static void do_recv(int node);
static void push_to_buffer(struct MP_msg_entry *msg);
static void Handler(int sig);

static int create_socket(void);
static int create_listen_socket(int *port);
static int accept_a_connection_from(int node, int listen_socket);
static int connect_to(int node, int base_port);
static void Dump_and_Quit(int sig);
static void Dump_Message_State(void);

#define TCP_PORT 5820
static int n_sends = 0, n_recvs = 0, n_sends_buffered = 0;
static int n_interupts = 0, n_sockets_serviced = 0;
static int sd[1024], sync_sd_prev, sync_sd_next;   /* socket descriptors */
static int salen = sizeof(struct sockaddr_in);
static struct sockaddr_in serv_addr;

#define MAX_MSGS 10000
static int idindex = 0;
static struct MP_msg_entry rmsg[1024], wmsg[1024], msg_null,
	                   *msg_list[MAX_MSGS];
static fd_set read_sds, write_sds;

#ifdef TRACE
void Trace_Dump();
extern int t_min, t_cur;
extern double t_0;
extern struct trace_element t_el[10000];
#endif

static sigset_t sigio_mask;

static void
Block_Signals(void)
{
    sigprocmask(SIG_BLOCK, &sigio_mask, NULL);
}

static void
Unblock_Signals(void)
{
    sigprocmask(SIG_UNBLOCK, &sigio_mask, NULL);
}

void
MP_Init(int *argc, char ***argv)
{
    FILE *fd;
    int iproc, i, j, n, r, flags, node, next, prev;
    int port = TCP_PORT, sync_port, sync_port_next, listen_socket;
    char *ptr, shortname[128];
    struct sigaction sigact;

    /*
     * Open .mplite.config and read in nprocs and hostnames.
     * Set nprocs and myproc, hostname[0..nprocs-1]
     */
    if (!(fd = fopen(".mplite.config", "r"))) {
        fprintf(stderr, "Couldn't open the .mplite.config file\n");
        exit(9);
    }
    fscanf(fd, "%d", &nprocs);

    gethostname(myhostname, sizeof(myhostname));
    if ((ptr = strchr(myhostname, '.')))
	*ptr = '\0';
    sprintf(logfilename, ".%s", myhostname);

#define LOG_TO_STDERR
#ifdef LOG_TO_STDERR
    mylog = stderr;
#else
    if (!(mylog = fopen(logfilename,"w")))
	errorxs(0, "Couldn't open the log file %s\n", logfilename);
#endif

    /*
     * Traditional format of the .mplite.config file:
     *   <numproc>
     *   0 <name>
     *   1 <name>
     *   ...
     * Alternative format of the .mplite.config file:
     *   <numproc>
     *   0 <name> <ipaddr>
     *   1 <name> <ipaddr>
     *   ...
     * Both work, the alternative one uses the <name> field to compare
     * against gethostname(), but uses the <ipaddr> (which could be a
     * symbolic name) for traffic.  The traditional one uses the same
     * <name> field for both pieces of information.
     */
    for (i=0; i<nprocs; i++) {
	char ipaddr[128];
        hostname[i] = malloc(128*sizeof(char));
        fscanf(fd, "%d %s", &iproc, hostname[i]);
	if ((ipaddr[0] = fgetc(fd)) != '\n')
	    fscanf(fd, "%s\n", ipaddr);
	else
	    ipaddr[0] = '\0';
	printf("hostname \"%s\"\n", hostname[i]);
	printf("ipaddr \"%s\"\n", ipaddr);
        strcpy(shortname, hostname[i]);
        if ((ptr = strchr(shortname, '.')))
	    *ptr = '\0';
        if (!strcmp(shortname, myhostname))
	    myproc = i;
	if (ipaddr[0])
	    strcpy(hostname[i], ipaddr);  /* but use this addr */
    }
    if (myproc < 0)
	errorxs(myproc, "%s had no entry in .mplite.config\n", myhostname);

    mp4("%s is node %d of %d\n", myhostname, myproc, nprocs);
    for (i=0; i<nprocs; i++)
	mp3("   %d ==> %s\n", i, hostname[i]);

    /* Set up a socket for each host.
     * Handshake to each, with a timeout.
     *    Report any errors to the master.
     *    If master is the problem, dump to .mplite.myhostname.
     */

    listen_socket = create_listen_socket(&port);
    sync_port = port+1;

    for (i=0; i<nprocs; i++) {
        if (i < myproc )    /* listen for the connection */
	    sd[i] = accept_a_connection_from(i, listen_socket);
	/*
	 * Wait for go signal from lower iprocs before connecting
         */
        if (i == myproc && myproc > 0)
	    if ((n = recv(sd[myproc-1], &j, sizeof(j), 0)) < 0)
		errorx(n, "tcp.c#MP_Init() - recv() go signal");

        if (i > myproc)  /* initiate the connection */
	    sd[i] = connect_to(i, TCP_PORT);
    }

    /* Send 'go' signal for the next iproc to connect */
    if (myproc < nprocs-1)
        if ((n = send(sd[myproc+1], &j, sizeof(j),0)) < 0)
            errorx(n, "tcp.c#MP_Init() - send() go signal");

    (void) close(listen_socket);

    /* Set up blocking sockets to the next and previous nodes for MP_Sync() */
    if (nprocs > 1) {
	mp1("\nSetting up special sockets for MP_Sync()\n");
	next = (myproc+1) % nprocs;
	prev = (myproc+nprocs-1) % nprocs;

	mp3("sync_port = %d (port = %d)\n", sync_port, port);
	listen_socket = create_listen_socket(&sync_port);
	/* trade sync_port values, up- and down-stream */
	for (r = -1; r <= 0; )
	    r = send(sd[prev], &sync_port, sizeof(int), 0);
	for (r = -1; r <= 0; )
	    r = recv(sd[next], &sync_port_next, sizeof(int), 0);
	if (myproc == 0) {
	    sync_sd_next = connect_to(next, sync_port_next);
	    if (nprocs == 2)
		sync_sd_prev = sync_sd_next;
	    else
		sync_sd_prev = accept_a_connection_from(prev, listen_socket);
	} else {
	    sync_sd_prev = accept_a_connection_from(prev, listen_socket);
	    if (nprocs == 2)
		sync_sd_next = sync_sd_prev;
	    else
		sync_sd_next = connect_to(next, sync_port_next);
	}
	mp2("%d sockets created for MP_Sync()\n", nprocs == 2 ? 1 : 2);
	(void) close(listen_socket);
    }
    MP_Sync();   /* Just to test it out */

#ifdef TRACE
    t_0 = MP_Time();
#endif

    /*
     * Cntl-c or kill -HUP will dump the message queue state then exit,
     * Handler() will handle SIGIO interupts
     */
    sigemptyset(&sigio_mask);
    sigaddset(&sigio_mask, SIGIO);  /* used to block sign during queue manip */
    sigact.sa_handler = Handler;
    sigact.sa_mask = sigio_mask;  /* block SIGIO during handlers */
    sigact.sa_flags = 0;  /* _not_ ONESHOT, e.g. */
    sigaction(SIGIO, &sigact, NULL);
    sigact.sa_handler = Dump_and_Quit;
    sigaction(SIGINT, &sigact, NULL);

    FD_ZERO(&read_sds);
    FD_ZERO(&write_sds);
    for (i=0; i<nprocs; i++)
	rmsg[i].next = wmsg[i].next = NULL;
    msg_null.stats = msg_null.nbytes = msg_null.nleft = msg_null.send_flag = 0;
    msg_null.src = msg_null.dest = msg_null.trace_id = -1;

    for (i=0; i<MAX_MSGS; i++)
	msg_list[i] = NULL;

    /* Enable asynchronous I/O now on all sockets */
    for (i=0; i<nprocs; i++) {
	if (i == myproc) continue;
	if (fcntl(sd[i], F_SETOWN, getpid()) < 0)
	    errorx(errno, "MP_Init: fcntl(SETOWN) failed");
	if ((flags = fcntl(sd[i], F_GETFL, 0)) < 0)
	    errorx(errno, "MP_Init: fcntl(GETFL) failed");
	dmp2("initial fcntl flags = 0x%x\n", flags);
	flags |= O_NONBLOCK | O_ASYNC;
	if (fcntl(sd[i], F_SETFL, flags) < 0)
	    errorx(errno, "MP_Init: fcntl(SETFL) failed");
	dmp2("final fcntl flags = 0x%x\n", flags);
    }

    /* Handshake with all, just to be polite */
    for (i=0; i<nprocs; i++)
	if (i != myproc)
	    MP_Send(&myproc, sizeof(int), i);

    for (i=0; i<nprocs; i++)
	if (i != myproc) {
	    MP_Recv(&node, sizeof(int), i);
	    if (node != i)
		mp3("WARNING - socket(%d) connected to node %d in MP_Init()\n",
		  i, node);
	}

    n_sends = n_recvs = n_interupts = n_sockets_serviced = 0;

    mp1("MP_Init: done with all the handshaking, exiting.\n");
    mp1("------------------------------------------------\n");
}

void
MP_Finalize(void)
{
    int i;

    MP_Sync();
    mp1("------------------------------------------------\n");
    for(i=0; i<nprocs; i++)
	if (i != myproc)
	    close(sd[i]);
    mp3("%d sends with %d partially buffered\n", n_sends, n_sends_buffered);
    mp2("%d receives\n", n_recvs);
    mp3("%d interupts handled servicing %d sockets\n",
      n_interupts, n_sockets_serviced);
    mp1("************************************************\n");
    fclose(mylog);
#if defined (TRACE)
    Trace_Dump();
#endif
}

/*
 * Handler()  - Handles SIGIO interupts from when data arrives or is sent
 *              on each socket.  Handler() services these by retrying the
 *              active sends and recvs on each socket until they are completed.
 * MP_Send()  - MP_ASend() then MP_Wait().
 * MP_ASend() - Asynchronous send - Make message header and initiate send.
 * MP_Recv()  - MP_ARecv() then MP_Wait().
 * MP_ARecv() - Asynchronous recv - Make message header and initiate recv.
 * MP_Wait()  - iRecv: block until full message has arrived.
 *            - iSend: Dump to send buffer if part of message is still left.
 */
void
MP_ASend(void *buf, int nbytes, int dest, int *msg_id)
{
    struct MP_msg_entry *msg, *mptr;

    Block_Signals();

    dmp3("MP_ASend(,%d,%d)\n", nbytes, dest);
#ifdef USER_FRIENDLY
    node_range_check(dest);
#endif

    while (msg_list[++idindex % MAX_MSGS]) ;
    *msg_id = idindex = idindex % MAX_MSGS;
    if (!(msg = msg_list[idindex] = malloc(sizeof(struct MP_msg_entry))))
	errorx(0, "tcp.c#MP_ASend() - malloc() failed");

    *msg = msg_null;
    msg->buf = msg->cur_ptr = buf;
    msg->nbytes = msg->nleft = nbytes;
    msg->dest = dest;
    msg->send_flag = 1;

#ifdef TRACE
    if (nbytes > t_min && ++t_cur < 10000) {
	msg->trace_id = t_cur;
	t_el[t_cur].node = dest;
	t_el[t_cur].nbytes = nbytes;
	t_el[t_cur].start = MP_Time();
    }
#endif

    /* Add to the end of the queue for this socket */
    for (mptr = &wmsg[dest]; mptr->next; mptr = mptr->next) ;
    mptr->next = msg;

    /* initiate the write by calling do_send() if sd is not active now */
    if (!FD_ISSET(sd[dest], &write_sds)) {
        FD_SET(sd[dest], &write_sds);
        do_send(dest);
    }

    Unblock_Signals();
    n_sends++;
    dmp3("MP_ASend(,%d,%d): exiting\n", nbytes, dest);
}

void
MP_Send(void *buf, int nbytes, int dest)
{
    int msg_id;

    MP_ASend(buf, nbytes, dest, &msg_id);
    MP_Wait(&msg_id);
}

void
MP_ARecv(void *buf, int nbytes, int src, int *msg_id)
{
    struct MP_msg_entry *mptr, *msg;

    Block_Signals();
    dmp3("MP_ARecv(,%d,%d)\n", nbytes, src);
#ifdef USER_FRIENDLY
    node_range_check(src);
#endif

    while (msg_list[++idindex%MAX_MSGS]) ;
    *msg_id = idindex = idindex % MAX_MSGS;
    if (!(msg = msg_list[idindex] = malloc(sizeof(struct MP_msg_entry))))
	errorx(0, "tcp.c#MP_ARecv() - malloc() failed");

    *msg = msg_null;
    msg->buf = msg->cur_ptr = buf;
    msg->nbytes = msg->nleft = nbytes;
    msg->src = src;

#ifdef TRACE
    if (nbytes > t_min && ++t_cur < 10000) {
	msg->trace_id = t_cur;
	t_el[t_cur].node = src;
	t_el[t_cur].nbytes = -nbytes;
	t_el[t_cur].start = MP_Time();
    }
#endif

    /* Add to the end of the recv queue for this socket */
    for (mptr = &rmsg[src]; mptr->next; mptr = mptr->next) ;
    mptr->next = msg;

    /* initiate the read by calling do_recv() if sd is not active now */
    if (!FD_ISSET(sd[src], &read_sds)) {
	FD_SET(sd[src], &read_sds);
	do_recv(src);
    }

    Unblock_Signals();
    ++n_recvs;
    dmp3("MP_Recv(,%d,%d): exiting\n", nbytes, src);
}

void
MP_Recv(void *buf, int nbytes, int src)
{
    int msg_id;

    MP_ARecv(buf, nbytes, src, &msg_id);
    MP_Wait(&msg_id);
}

void
MP_Wait(int *msg_id)
{
    int counter = 0;
    double tmin = PANIC_TIME, t, t0 = 0.0;
    struct MP_msg_entry *msg;

    if (!(msg = msg_list[*msg_id]))
	errorx(*msg_id, "MP_Wait() has NULL msg");

    if (msg->nleft == 0)
	;
    else if (msg->send_flag == 1) {
	Block_Signals();
        dmp1("MP_Wait() push rest to send buffer\n");
        msg->send_flag = 2;
        push_to_buffer(msg);
        Unblock_Signals();
    } else {
	dmp1("MP_Wait() going into a wait loop\n");
        while (msg->nleft) {
	    if (++counter > 200000) {
		if (t0 == 0.)
		    t0 = MP_Time();
		t = MP_Time() - t0;
		if (t > tmin) {  /* log any long waits */
		    mp3("MP_Wait(%d) blocking for %d seconds\n",
		      msg->nbytes, (int) t);
		    Dump_Message_State();
		    tmin *= 2;
		}
	    }
	}
	dmp1("MP_Wait() wait loop done\n");
    }

#ifdef TRACE
    if (msg->trace_id > 0 && msg->trace_id < 10000) {
	t_el[msg->trace_id].end = MP_Time();
        t_el[msg->trace_id].npackets = msg->stats;
        if (msg->send_flag > 1)
	    t_el[msg->trace_id].node += nprocs; /* signifying a buffered send */
    }
#endif

/*   free(msg);*/
    free(msg_list[*msg_id]);
    msg_list[*msg_id] = NULL;
/*   mp2("MP_Wait() msg_id = %d retired\n", *msg_id);*/

    dmp1("MP_Wait: exiting\n");
}

int
MP_Nprocs(int *n)
{ 
    *n = nprocs; 
    return nprocs;
}

int
MP_Myproc(int *m)
{
    *m = myproc;
    return myproc;
}

static void
do_send(int node)
{
    int n;
    struct MP_msg_entry *msg;

    msg = wmsg[node].next;
    dmp3("do_send: sending %d bytes to %d --> ", msg->nleft, msg->dest);
    n = send(sd[node], msg->cur_ptr, msg->nleft, 0);
    if (n < 0 && errno == EWOULDBLOCK)
	n = 0;
    dmp2("%d bytes actually sent\n", n);
    if (n < 0)
	errorx(n,"tcp.c#do_send() - send()");
    msg->nleft -= n;
    msg->cur_ptr += n;
    msg->stats++;

    if (msg->nleft == 0) {
	wmsg[node].next = msg->next;   /* detach msg header from the queue */
        if (msg->send_flag == 3) {     /* free buffer and msg header */
            free(msg->buf);
            free(msg);
        }
        if (wmsg[node].next == NULL)
	    FD_CLR(sd[node], &write_sds);
        else
	    do_send(node);
    }
}

static void
do_recv(int node)
{
    int n;
    struct MP_msg_entry *msg;

    msg = rmsg[node].next;
    dmp3("do_recv: reading %d bytes from %d <-- ", msg->nleft, msg->src);
    n = recv(sd[node], msg->cur_ptr, msg->nleft, 0);
    if (n < 0 && errno == EWOULDBLOCK) n = 0;
    dmp2("%d bytes actually read\n", n);
    if (n < 0)
	errorx(errno,"tcp.c#do_recv() - recv()");
    msg->nleft -= n;
    msg->cur_ptr += n;
    msg->stats++;

    if (msg->nleft == 0) {
        rmsg[node].next = msg->next;   /* detach msg header from the queue */
        if (rmsg[node].next == NULL)
	    FD_CLR(sd[node], &read_sds);
        else
	    do_recv(node);
    }
}

static void
push_to_buffer(struct MP_msg_entry *msg)
{
   int node;
   struct MP_msg_entry *new_msg, *msg_ptr;
   void* send_buffer;

   if( msg->nleft == 0 ) return;   /* just incase the send completed */

   node = msg->dest;
   msg_ptr = &wmsg[node];
   while( msg_ptr->next != msg ) {
      msg_ptr = msg_ptr->next;
   }

   send_buffer = malloc(msg->nleft);
   new_msg = malloc(sizeof(struct MP_msg_entry));
   if( ! send_buffer || ! new_msg )
       errorx(msg->nleft,"tcp.c#push_to_buffer() - malloc(send buffer)");
   *new_msg = *msg;
   memcpy(send_buffer, msg->cur_ptr, msg->nleft);
   new_msg->buf = new_msg->cur_ptr = send_buffer;
   new_msg->send_flag = 3;     /* signifying a buffered send */
   msg_ptr->next = new_msg;  new_msg->next = msg->next;
   msg->nleft = 0;   /* signifying that old header is free to reuse */
   n_sends_buffered++;
   dmp2("Pushed %d bytes into a send buffer\n",new_msg->nleft);
}

/*
 * SIGIO handler for asynchronous I/O of sends and receives
 */
static void
Handler(int sig)
{
    int node, r;
    fd_set rsds, wsds;
    struct timeval timeout;

    dmp1("*** Handler() entered from a SIGIO\n");
    rsds = read_sds;
    wsds = write_sds;

/* #ifndef(linux)  \* select() doesn't work under Linux (Redhat 4.2) */
    timeout.tv_sec = timeout.tv_usec = 0;
    if ((r = select(OPEN_MAX, &rsds, &wsds, NULL, &timeout)) < 0)
	errorx(errno, "Handler: select");
    if (r == 0)
	goto out;  /* sigio will happen even when we "actually" send */
/* #endif */

    for (node=0; node<nprocs; node++) {
	if (node == myproc) continue;
	if (FD_ISSET(sd[node], &wsds)) {
            ++n_sockets_serviced;
            do_send(node);
        }

        if (FD_ISSET(sd[node], &rsds)) {
            ++n_sockets_serviced;
            do_recv(node);
        }
    }

out:
    ++n_interupts;
    dmp1("Handler: return\n");
}

/* Barrier sync using separate blocking sockets
 *   - this avoid conflicts with the existing iRecvs so you can for example
 *     post an iRecv and do an MP_Sync() before sending the data to insure
 *     that the iRecv is preposted and the data will bypass the send buffer.
 *   - This is not the fastest sync(), but it is a simple and sure method.
 *   - Must protect agains interupts so put send/recv in while loops.
 */

void
MP_Sync(void)
{
    int r, token = 777;

    if (nprocs == 1)
	return;
    dmp1("MP_Sync() started - ");

#ifdef TRACE
    if (++t_cur < 10000) {
        t_el[t_cur].nbytes = 0;
        t_el[t_cur].start = MP_Time();
    }
#endif

    if (myproc == 0)
	for (r = -1; r <= 0; )
	    r = send(sync_sd_next, &token, sizeof(int), 0);
    for (r = -1; r <= 0; )
	r = recv(sync_sd_prev, &token, sizeof(int), 0);
    for (r = -1; r <= 0; )
	r = send(sync_sd_next, &token, sizeof(int), 0);
    for (r = -1; r <= 0; )
	r = recv(sync_sd_prev, &token, sizeof(int), 0);
    if (myproc != 0)
	for (r = -1; r <= 0; )
	    r = send(sync_sd_next, &token, sizeof(int), 0);

#ifdef TRACE
    if (t_cur < 10000)
	t_el[t_cur].end = MP_Time();
#endif
    dmp1("done\n");
}

/* Utility routines for above */

static int
create_socket(void)
{
    int sockdr, one = 1, buffer_orig = 0, d = sizeof(int), r;
    static int print_once = 1;

    if ((sockdr = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	errorx(errno, "MP_Init()-create_socket()");

    /* Set TCP_NODELAY */
    if (setsockopt(sockdr, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one)) < 0)
	errorx(errno, "MP_Init()-setsockopt(NODELAY)");

    /* Attempt to set the send and receive buffer sizes to 4 MB */
    if (getsockopt(sockdr, SOL_SOCKET, SO_RCVBUF, (char *) &buffer_orig,
      (void *) &d) < 0)
	errorw(errno, "MP_Init()-getsockopt(SO_RCVBUF)");

    r = -1;
    while (r < 0 && buffer_size > buffer_orig) {
        r = setsockopt(sockdr, SOL_SOCKET, SO_SNDBUF,
          (char *) &buffer_size, sizeof(buffer_size));
        r = setsockopt(sockdr, SOL_SOCKET, SO_RCVBUF,
          (char *) &buffer_size, sizeof(buffer_size));
        if (r < 0)
	    buffer_size /= 2;
    }

    if (getsockopt(sockdr, SOL_SOCKET, SO_RCVBUF, (char *) &buffer_size,
      (void *) &d) < 0)
	errorw(errno, "MP_Init()-getsockopt(SO_RCVBUF) #2");
    if (print_once) {
        mp3("Send/Receive Buffers: %d ==> %d bytes\n",
	  buffer_orig, buffer_size);
        print_once = 0;
    }
    return sockdr;
}

static int
create_listen_socket(int *port)
{
   int sockdr, bound = -2, tries = 0;

   sockdr = create_socket();
   serv_addr.sin_family      = AF_INET;
   serv_addr.sin_port        = htons(*port);
   serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   while( bound < 0 && ++tries < 10) {
      dmp1("Trying to bind for listen_socket\n");
      bound = bind( sockdr, (struct sockaddr *) &serv_addr, salen);
      if( bound < 0 ) {
         mp3("Can't bind to listen port %d - errno = %d\n", *port,errno);
         serv_addr.sin_port   = htons(++ *port);
      }
   }
   if( bound < 0 ) errorx(errno,"MP_Init()-bind()");
   mp2("Bind to listen port %d done - ", *port);

   listen(sockdr, 5);
   mp1("listen completed\n");

   return sockdr;
}

static int
accept_a_connection_from(int node, int listen_socket)
{
    int sockdr;

    mp2("Trying to accept a connection from %d - ", node);
    bzero((char *)&serv_addr, sizeof(serv_addr));
    if ((sockdr = accept(listen_socket, (struct sockaddr *) &serv_addr,
      (void *) &salen )) < 0)
	errorx(errno,"MP_Init()-accept()");
    mp1("accepted\n");
    return sockdr;
}

static int
connect_to(int node, int base_port)
{
    struct hostent *addr;
    int sockdr, port, connected = -2, tries = 0;

    mp2("Trying to connect to node %d - ", node);
    if (!(addr = gethostbyname(hostname[node])))
	errorxs(0, "Invalid hostname %s", hostname[node]);

   /* Try to connect for up to 60 seconds before giving up */
    while (connected < 0 && tries++ < 300) {
	serv_addr.sin_family = addr->h_addrtype;
        bcopy(addr->h_addr, &serv_addr.sin_addr.s_addr, addr->h_length);
        port = base_port + (tries-1) % 10;
        serv_addr.sin_port = htons(port);

        /* Alpha Linux needs the socket reset after each try
	 * for some reason
         */
        sockdr = create_socket();
        if ((connected = connect(sockdr, (struct sockaddr *) &serv_addr,
	  salen)) < 0) {
	    (void) close(sockdr);
	    if (errno != ECONNREFUSED)
		errorx(errno,"MP_Init()-connect()");
	    if (!(tries%10)) {
		printf("Sleeping 2 secs after trying ports %d-%d for node %d\n",
                  TCP_PORT, TCP_PORT+8, node);
		sleep(2);
	    }
	}
    }
    if (connected < 0)
	errorx(errno, "MP_Init()-connect()");
    mp3("connected after %d tries on port %d\n", tries, port);
    return sockdr;
}

static void
Dump_and_Quit(int sig)
{
   mp1("\n\nReceived a SIGINT (cntl-c) --> dump and exit\n");
   Dump_Message_State();
   fclose(mylog);
   exit(-1);
}

static void
Dump_Message_State(void)
{
   int node;
   struct MP_msg_entry *msg;

   mp1("\nDumping the current state of the message queues\n");
   for( node=0; node<nprocs; node++) if( node != myproc ) {
      msg = wmsg[node].next;
      while( msg != NULL ) {
         mp5("  --> Sent %d of %d bytes to node %d %s\n", 
             msg->nbytes - msg->nleft, msg->nbytes, node,
             msg->send_flag == 3 ? "- the rest is buffered" : "");
         msg = msg->next;
   }  }
   for( node=0; node<nprocs; node++) if( node != myproc ) {
      msg = rmsg[node].next;
      while( msg != NULL ) {
         mp4("  <-- Received %d of %d bytes from node %d\n", 
             msg->nbytes - msg->nleft, msg->nbytes, node);
         msg = msg->next;
   }  }
   mp1("\n");
}

