/*
 * Timing, operation counting, and memory tracking
 * Code by Dave Turner in 1996 (Based on original work by David Deaven)
 */

#include "mplite.h"   /* Globals myproc, nprocs, mylog, hostname, logfile */
#include "mpfuncs.h"

/*#undef ELPRINT*/
#define MPRINTF if(myproc==0)printf

/*
  void MP_Enter(char *name);
  void MP_Leave(double nflops);
  void MP_Time_Report(char *filename);
  void MP_Reset_Time();
  void MEMORY(char *s, int nbytes);
  void *mymalloc(size_t nbytes);
  void *myremalloc(void *ptr, int nbytes);
  void *myfree(void *ptr);
  void MEMSTRING(char *s);
*/


/* Report significant changes in allocated memory  */

void MEMORY(char *s, int nbytes)
{
   static double total_mb = 0.0;
   double mb;

   mb = (double) nbytes / (1024.0 * 1024.0);
   total_mb += mb;
   MPRINTF("MEMORY  %-10s  %8.3f MB  --> %8.3f MBytes\n",s,mb,total_mb);
}

struct memblock {
   void *ptr;           /* Pointer to the memory */
   size_t size;           /* Number of bytes */
   char name[20];       /* Name of array */
   void *next;          /* Pointer to the next memblock in the linked list */
};

static struct memblock *membase = NULL;
static char mem_string[20] = "\0";
static double total_mb = 0.0, max_mb = 0.0;

void MEMSTRING(char *s)
{ strcpy(mem_string,s); }

void memdump(struct memblock *b)
{
   double mb;

   mb = ((double) ((int) b->size) ) / (1024.0 * 1024.0);
   total_mb += mb;
   if( total_mb > max_mb) {
      max_mb = total_mb;
      if( fabs(mb) >= 0.1 )
         MPRINTF("-->MEMORY  %-10s  %8.3f MB  --> %8.3f MBytes\n",
                  b->name, mb, total_mb);
   }
}  

void *mymalloc(size_t nbytes)
{
   struct memblock *b;

   if( membase == NULL ) {
      b = membase = malloc(sizeof(struct memblock));
   } else {
      b = membase;
      while( b->next != NULL ) b = b->next;
      b->next = malloc(sizeof(struct memblock));
      b = b->next;
   }
   b->ptr = malloc(nbytes);
   b->size = nbytes;
   b->next = NULL;
   strcpy(b->name,mem_string);
   strcpy(mem_string,"\0");

   memdump(b);
   return b->ptr;
}

void myfree(void *ptr)
{
   struct memblock *b, *blast;

   for( b = membase, blast = NULL; b != NULL; b = b->next) {
      if( b->ptr == ptr) break;
      blast = b;
   }
   if( b == NULL ) {
      MPRINTF("-->MYFREE - Trying to free a non-existent block\n");
   } else {
      b->size = - b->size;
      memdump(b);
      if( blast == NULL ) membase = b->next;
      else blast->next = b->next;
      free(ptr);
   }
} 

void *myrealloc(void *ptr, size_t nbytes)
{
   struct memblock *b, *blast;
   void *ptr_new;

   for( b = membase, blast = NULL; b != NULL; b = b->next) {
      if( b->ptr == ptr) break;
      blast = b;
   }
   if( b == NULL || ptr == NULL) {
      ptr_new = mymalloc(nbytes);
   } else {
      b->size = nbytes - b->size;
      memdump(b);
      ptr_new = b->ptr = realloc(ptr, nbytes);
      b->size = nbytes;
   }
   return ptr_new;
}


/* Timing routines for the entire code */

struct timed_block {
   char name[40];          /* name from input parameters */
   double t_enter;         /* when did entry occur */
   double total_time;      /* total time spent in routine and children */
   double child_time;      /* total time spent in lower routines */
   int ncalls;            /* number of entries into this block of code */
   double nflops;          /* total number of floating-point operations */
};

#define NUM_TIMED_BLOCKS 500
static struct timed_block block[NUM_TIMED_BLOCKS];
static int stack[NUM_TIMED_BLOCKS], max_block = 0, stack_ptr = 0, id;

/* Reset the timed blocks after step 0 */
void MP_Reset_Time(void)
{
   double t = MP_Time();
   int i;

   for(i=1;i<NUM_TIMED_BLOCKS;i++) {         /* dont reset block 0 (all) (init block) */
      block[i].t_enter = t;     /* reset t_enter on main-loop */
      block[i].total_time = 0;  /* zero out the total times */
      block[i].child_time = 0;  /* zero out the child times */
      block[i].ncalls = 0;      /* zero out the number of calls */
      block[i].nflops = 0;      /* zero out the nflops */
   }
}



char* Cleanup_String(char *string)
{
   char *p, *new_string;

/* Linux won't allow modification of the constant? string,
 *    so make a copy first */

   new_string = malloc(strlen(string)+1);
   strcpy(new_string, string);   

   if( strlen(new_string) > 39 ) new_string[39] = '\0';

/*   p = &string[strlen(string)-1];*/
/*   while( *p == ' ' && strlen(string) > 0 ) { *p = '\0'; p--; }*/

   /* Treat a pipe '|' as a string termination */

   p = strstr(new_string, "|");
   if( p != NULL ) *p = '\0' ;

   return new_string;
}


/* Enter a timed block
 *   - f77 doesn't always terminate strings properly, so it's possible
 *     that some garbage will be tacked onto the end of the 'name' string.
 *     If you encounter problems, use a pipe '|' at the end of the string
 *     to manually terminate it.
 */

void MP_Enter(char *name)
{
   double t = MP_Time();
   int i;

#ifdef ELPRINT
   MP_Sync();
   MPRINTF("-->ENTER %s\n",name);
#endif

   /* Find the block id or add a new one if necessary */

/*   printf("MP_Enter(%s)\n",name);*/
   name = Cleanup_String(name);
/*   printf("MP_Enter(%s)\n",name);*/
   for(i=0, id = -1; i<max_block; i++)
     if(strcmp(name, block[i].name) == 0) id = i;
   if(id == -1) {
      id = max_block++;
      if( id > NUM_TIMED_BLOCKS ) {
         MPRINTF("ERROR - Increase the number of timed blocks in timer.c\n");
         exit(1);
      }
      strcpy(block[id].name, name);
      block[id].total_time = 0;
      block[id].child_time = 0;
      block[id].ncalls = 0;
      block[id].nflops = 0;
   }

   /* Add it to the stack */
   stack[stack_ptr++] = id;
   block[id].t_enter = t;
   block[id].ncalls++;
}

/* Leave a timed block, reverting to the previous context.  */
void MP_Leave(double nflops)
{
   double dt, t = MP_Time();

   if( stack_ptr <= 0 ) {
      fprintf(stderr,"WARNING - MP_Leave() has no matching MP_Enter()\n");
      return;
   }
   id = stack[--stack_ptr];

   dt = t - block[id].t_enter;
   /* Add this time to the current time block and to the parent's child_time */
   block[id].total_time += dt;
   if(stack_ptr > 0) block[stack[stack_ptr-1]].child_time += dt;

   /* These are the best checks I could come up with to keep out ints, etc */

/*   if( !isnan(nflops) && nflops < 1.e16 && nflops > -1.e10 ) */
/*   if( !isnan(nflops) && *(int *) &nflops != 0 ) */
   if( finite(nflops) ) {
/*      printf("(double) nflops = %g\n", nflops);*/
/*      printf("(int) nflops = %d\n", nflops);*/
/*      printf("(int) nflops = %d\n", *(int *) &nflops);*/
/*      if( nflops >= 1.e0 ) */
         block[id].nflops += nflops;
   }

#ifdef ELPRINT
   MP_Sync();
   MPRINTF("<--leave %s\n",block[id].name);
#endif
}

/* f77 doesn't always properly terminate strings, so from f77 I use a
 * pipe '|' to terminate for this function
 */

double MP_Get_Time(char *name)
{
   int id;
   double time = 0.0;

   name = Cleanup_String(name);
/*   printf("MP_Get_Time(%s)\n",name);*/
   for(id=0; id<max_block; id++) {
      if( !strcmp(block[id].name,name) ) {
         time = block[id].total_time;
      }
   }
   return time;
}

double MP_Get_Flops(char *name)
{
   int id;
   double flops = 0.0;

   name = Cleanup_String(name);
/*   printf("MP_Get_Flops(%s)\n",name);*/
   for(id=0; id<max_block; id++) {
      if( !strcmp(block[id].name,name) ) {
         flops = block[id].nflops;
      }
   }
   return flops;
}

double MP_Myflops()
{
   int id;
   double myflops = 0.0;

   for(id=0; id<max_block; id++) {
      if( block[id].nflops > 0.0 ) {
         myflops += block[id].nflops;
      }
   }
   return myflops;
}

void MP_Time_Report(char *timefile)
{
   double dt, frac, total_time = 0, total_nflops = 0;
   int jd;
   FILE *fp;

   timefile = Cleanup_String(timefile);
   total_time = block[0].total_time;

      /* Sort by descending time spent in the block */

   for(id=0; id<max_block-1; id++) {
      dt = block[id].total_time - block[id].child_time;
      for(jd=id+1; jd<max_block; jd++) {
         if( block[jd].total_time - block[jd].child_time > dt ) {
            dt = block[jd].total_time - block[jd].child_time;
            block[max_block] = block[id];
            block[id] = block[jd];
            block[jd] = block[max_block];
   }   }   }

   if( myproc==0 ) {
      if( stack_ptr != 0 ) {
        fprintf(stderr,"Time_Report() called before timed blocks completed)\n");
        return;
      }
      fp = fopen(timefile,"w");
   }
   for(id=0; id<max_block; id++) {
      if( block[id].nflops > 0 ) total_nflops += block[id].nflops;
   }
   MP_dSum(&total_nflops,1);

   if( myproc==0 ) {
      if( total_time < 1.0e-6 ) {
          fprintf(fp,"Total time is zero for some reason\n");
          return;
      }

      fprintf(fp,"    time in seconds       +children    calls   name\n");
      for(id=0; id<max_block; id++) {
         dt = block[id].total_time - block[id].child_time;
         frac = 100 * dt / total_time;
         if(frac < 0.001) break;
         fprintf(fp, " %10.3f (%5.1f%%) %13.3f %8d  %-17.17s", 
            dt, frac, block[id].total_time, 
            block[id].ncalls, block[id].name);
         if(block[id].nflops == 0) fprintf(fp,"\n");
         else if( block[id].nflops < 0 )
            fprintf(fp," %9.2f MB/sec\n",-block[id].nflops * 1.0e-6 / dt);
         else fprintf(fp," %9.2f MFlops\n",block[id].nflops * 1.0e-6 / dt);
      }
      if( total_nflops > 0 )
         fprintf(fp, "\n %9.2f billion operations using %d processors "
              "--> %9.2f MFlops\n",
              total_nflops * 1.0e-9, nprocs, total_nflops *1.0e-6 / total_time);
      fclose(fp);
   }
}

